/*!
 * \file
 * \ingroup event_handle
 * \brief global event handling.
 *
 * \note Struct marking, and the variables adding_mark, mark_x, mark_y, max_mark and marks have been
 *       moved to mapwin.h
 */
#ifndef __EVENTS_H__
#define __EVENTS_H__

/*!
 * \name    Event types
 * @{
 */
enum { 
    EVENT_MOVEMENT_TIMER, /*!< event caused by the timer thread */
    EVENT_UPDATE_CAMERA, /*!< camera update event */
    EVENT_ANIMATE_ACTORS  /*!< animate actor event */
};
/*! @} */

extern int shift_on; /*!< flag indicating whether the Shift key is pressed. */
extern int alt_on; /*!< flag indicating whether the Alt key is pressed. */
extern int ctrl_on; /*!< flag indicating whether the Ctrl key is pressd. */

/*!
 * \ingroup event_handle
 * \brief   gets called when a use mouse click occurs in the quickbar.
 *
 *      Gets called when a use mouse click occurs in the quickbar.
 *
 * \param use_id    the id of the item to use
 */
void	quick_use(int use_id);

/*!
 * \ingroup event_handle
 * \brief       main event handling routine
 *
 *      Main event handling routine.
 *
 * \param event     a pointer to last event occurred.
 * \retval int
 * \callgraph
 */
int HandleEvent(SDL_Event *event);

#endif	// __EVENTS_H__
